import collections

from jinja2 import Template


class VCSConfiguration(object):
    def __init__(self, name, options, global_variables, special_variables, commit_message=None, finish_release=True):
        self.name = name

        if commit_message is None:
            commit_message = "Version updated {{ current_version }} -> {{ new_version }}"

        commit_message_template = Template(commit_message)

        template_variables = {}
        template_variables.update(global_variables)
        template_variables.update(special_variables)

        self.commit_message = commit_message_template.render(**template_variables)
        self.finish_release = finish_release

        self.options = {}
        for key, value in options.items():
            if isinstance(value, collections.Sequence):
                value_template = Template(value)
                self.options[key] = value_template.render(**template_variables)
            else:
                self.options[key] = value

        self.options.update(special_variables)

    @classmethod
    def from_dict(cls, vcs_configuration_dict, global_variables, special_variables):
        return VCSConfiguration(vcs_configuration_dict['name'],
                                vcs_configuration_dict.get('options', {}),
                                global_variables,
                                special_variables,
                                vcs_configuration_dict.get('commit_message', None),
                                vcs_configuration_dict.get('finish_release', True))



import six

import pytest

import tempfile
import shutil
import os
import subprocess


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true",
                     help="run slow tests")


def pytest_runtest_setup(item):
    if 'slow' in item.keywords and not item.config.getvalue("runslow"):
        pytest.skip("need --runslow option to run")


@pytest.fixture
def temp_empty_dir(request):
    tempdir = tempfile.mkdtemp()

    # Make sure the directory is empty, there are some issues on Travis CI
    shutil.rmtree(tempdir, ignore_errors=True)
    os.mkdir(tempdir)

    def fin():
        shutil.rmtree(tempdir, ignore_errors=True)

    request.addfinalizer(fin)
    return tempdir


@pytest.fixture
def safe_devnull():
    if six.PY2:
        devnull = open(os.devnull, 'w')

        def fin():
            devnull.close()

    else:
        devnull = subprocess.DEVNULL

    return devnull



import subprocess

import os
import pytest
from punch import vcs_configuration as vc
from punch.vcs_repositories import git_flow_repo as gfr, exceptions as re

pytestmark = pytest.mark.slow


@pytest.fixture
def temp_empty_git_dir(temp_empty_dir):
    subprocess.check_call(["git", "init", "-q", temp_empty_dir])
    subprocess.check_call(["git", "config", "user.email", "py.test@email.com"], cwd=temp_empty_dir)
    subprocess.check_call(["git", "config", "user.name", "PyTest"], cwd=temp_empty_dir)

    command_line = ["git", "flow", "init", "-d"]
    p = subprocess.Popen(command_line, cwd=temp_empty_dir, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    stdout, stderr = p.communicate()

    if p.returncode != 0:
        _error_text = "An error occurred executing '{}': {}\nProcess output was: {}"
        _error_message = _error_text.format(" ".join(command_line),
                                            stderr.decode('utf8'), stdout.decode('utf8'))

        raise ValueError(_error_message)

    return temp_empty_dir


@pytest.fixture
def temp_gitflow_dir(temp_empty_git_dir, safe_devnull):
    with open(os.path.join(temp_empty_git_dir, "README.md"), "w") as f:
        f.writelines(["# Test file", "This is just a test file for punch"])

    subprocess.check_call(["git", "add", "README.md"], cwd=temp_empty_git_dir, stdout=safe_devnull)
    subprocess.check_call(["git", "commit", "-m", "Initial addition"], cwd=temp_empty_git_dir,
                     stdout=safe_devnull)

    return temp_empty_git_dir


@pytest.fixture
def empty_vcs_configuration():
    return vc.VCSConfiguration('git', {}, {}, {'current_version': 'a', 'new_version': 'b'})

def test_init(temp_empty_git_dir, empty_vcs_configuration):
    repo = gfr.GitFlowRepo(temp_empty_git_dir, empty_vcs_configuration)

    assert repo.working_path == temp_empty_git_dir


def test_get_current_branch(temp_gitflow_dir, empty_vcs_configuration):
    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)
    assert repo.get_current_branch() == 'develop'


def test_get_tags(temp_gitflow_dir, empty_vcs_configuration):
    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)
    assert repo.get_tags() == ''


def test_init_with_uninitialized_dir(temp_empty_dir, empty_vcs_configuration):
    with pytest.raises(re.RepositorySystemError) as exc:
        gfr.GitFlowRepo(temp_empty_dir, empty_vcs_configuration)

    assert str(exc.value) == "The current directory {} is not a Git repository".format(temp_empty_dir)


def test_pre_start_release(temp_gitflow_dir, empty_vcs_configuration):
    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)
    repo.pre_start_release()


def test_pre_start_release_starting_from_different_branch(temp_gitflow_dir, safe_devnull, empty_vcs_configuration):
    subprocess.check_call(["git", "checkout", "-b", "new_branch"], cwd=temp_gitflow_dir, stdout=safe_devnull,
                     stderr=safe_devnull)

    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)
    repo.pre_start_release()

    assert repo.get_current_branch() == 'develop'


def test_pre_start_release_starting_from_different_branch_with_unstaged_changes(temp_gitflow_dir, safe_devnull, empty_vcs_configuration):
    subprocess.check_call(["git", "checkout", "-b", "new_branch"], cwd=temp_gitflow_dir, stdout=safe_devnull,
                     stderr=safe_devnull)
    with open(os.path.join(temp_gitflow_dir, "README.md"), "w") as f:
        f.writelines(["Unstaged lines"])

    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)

    repo.pre_start_release()

    assert repo.get_current_branch() == 'develop'


def test_pre_start_release_starting_from_different_branch_with_uncommitted_changes(temp_gitflow_dir, safe_devnull, empty_vcs_configuration):
    subprocess.check_call(["git", "checkout", "-b", "new_branch"], cwd=temp_gitflow_dir, stdout=safe_devnull,
                     stderr=safe_devnull)
    with open(os.path.join(temp_gitflow_dir, "README.md"), "w") as f:
        f.writelines(["Unstaged lines"])
    subprocess.check_call(["git", "add", "README.md"], cwd=temp_gitflow_dir, stdout=safe_devnull,
                     stderr=safe_devnull)

    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)
    with pytest.raises(gfr.RepositoryStatusError):
        repo.pre_start_release()


def test_start_release(temp_gitflow_dir, empty_vcs_configuration):
    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)
    repo.pre_start_release()
    repo.start_release()
    assert repo.get_current_branch() == "release/b"


def test_finish_release_without_changes(temp_gitflow_dir, empty_vcs_configuration):
    release_name = empty_vcs_configuration.options['new_version']
    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)
    repo.pre_start_release()
    repo.start_release()
    repo.finish_release()
    assert repo.get_current_branch() == "develop"
    assert release_name not in repo.get_tags()


def test_finish_release_with_changes(temp_gitflow_dir, empty_vcs_configuration):
    release_name = empty_vcs_configuration.options['new_version']

    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)
    repo.pre_start_release()
    repo.start_release()

    with open(os.path.join(temp_gitflow_dir, "version.txt"), "w") as f:
        f.writelines([release_name])

    repo.finish_release()
    assert repo.get_current_branch() == "develop"
    assert release_name in repo.get_tags()


def test_tag(temp_gitflow_dir, empty_vcs_configuration):
    repo = gfr.GitFlowRepo(temp_gitflow_dir, empty_vcs_configuration)

    repo.tag("just_a_tag")

    assert "just_a_tag" in repo.get_tags()



import os
import six

from punch import replacer


class FileUpdater(object):
    def __init__(self, file_configuration):
        self.file_configuration = file_configuration
        self.rep = replacer.Replacer(file_configuration.config['serializer'])

    def get_summary(self, current_version, new_version):
        return self.rep.run_all_serializers(current_version, new_version)

    def update(self, current_version, new_version):
        if not os.path.exists(self.file_configuration.path):
            if six.PY2:
                raise IOError("The file {} does not exist".format(self.file_configuration.path))
            else:
                raise FileNotFoundError("The file {} does not exist".format(self.file_configuration.path))

        with open(self.file_configuration.path, 'r') as f:
            old_file_content = f.read()

        new_file_content = self.rep.replace(old_file_content, current_version, new_version)

        with open(self.file_configuration.path, 'w') as f:
            f.write(new_file_content)



import os
import six
from punch.vcs_repositories import exceptions as e
from punch.vcs_repositories import vcs_repo as vr


class GitRepo(vr.VCSRepo):
    def __init__(self, working_path, config_obj):
        if six.PY2:
            super(GitRepo, self).__init__(working_path, config_obj)
        else:
            super().__init__(working_path, config_obj)

        self.make_release_branch = self.config_obj.options.get('make_release_branch', True)

    def _check_config(self):
        # Tag names cannot contain spaces
        tag = self.config_obj.options.get('tag', '')
        if ' ' in tag:
            raise e.RepositoryConfigurationError(
                """You specified "'tag': {}". Tag names cannot contain spaces""".format(tag))

    def _check_system(self):
        if six.PY2:
            super(GitRepo, self)._check_system()
        else:
            super()._check_system()

        if not os.path.exists(os.path.join(self.working_path, '.git')):
            raise e.RepositorySystemError("The current directory {} is not a Git repository".format(self.working_path))

    def _set_command(self):
        self.commands = ['git']
        self.command = 'git'

    def get_current_branch(self):
        stdout = self._run([self.command, "rev-parse", "--abbrev-ref", "HEAD"])

        branch = stdout.replace("\n", "")

        return branch

    def get_tags(self):
        return self._run([self.command, "tag"])

    def get_branches(self):
        return self._run([self.command, "branch"])

    def pre_start_release(self):
        output = self._run([self.command, "status"])
        if "Changes to be committed:" in output:
            raise e.RepositoryStatusError("Cannot checkout master while repository contains uncommitted changes")

        self._run([self.command, "checkout", "master"])

        branch = self.get_current_branch()

        if branch != "master":
            raise e.RepositoryStatusError("Current branch shall be master but is {}".format(branch))

    def start_release(self):
        if self.make_release_branch:
            self._run([self.command, "checkout", "-b", self.config_obj.options['new_version']])

    def finish_release(self):
        branch = self.get_current_branch()

        self._run([self.command, "add", "."])

        output = self._run([self.command, "status"])

        if "nothing to commit, working directory clean" in output and self.make_release_branch:
            self._run([self.command, "checkout", "master"])
            self._run([self.command, "branch", "-d", branch])
            return

        command_line = [self.command, "commit"]
        command_line.extend(["-m", self.config_obj.commit_message])

        self._run(command_line)

        if self.make_release_branch:
            self._run([self.command, "checkout", "master"])
            self._run([self.command, "merge", branch])
            self._run([self.command, "branch", "-d", branch])

        try:
            tag_value = self.config_obj.options['tag']
        except KeyError:
            tag_value = self.config_obj.options['new_version']

        if self.config_obj.options.get('annotate_tags', False):
            annotation_message = self.config_obj.options.get('annotation_message', "Version {{ new_version }}")
            self._run([self.command, "tag", "-a", tag_value, "-m", annotation_message])
        else:
            self._run([self.command, "tag", tag_value])

    def post_finish_release(self):
        pass

    def tag(self, tag_name):
        self._run([self.command, "tag", tag_name])



import mock
from punch.vcs_use_cases import use_case


def test_delegate_all():
    repo = mock.Mock()
    uc = use_case.VCSUseCase(repo)

    uc.some_function("unnamed_parameter", named_parameter="named_parameter")

    assert repo.some_function.called_with("unnamed_parameter", named_parameter="named_parameter")


def test_delegate_override():
    repo = mock.Mock()

    class MyVCSUseCase(use_case.VCSUseCase):
        def some_function(self, *args, **kwds):
            pass

    uc = MyVCSUseCase(repo)
    uc.some_function("unnamed_parameter", named_parameter="named_parameter")

    assert not repo.some_function.called



import collections
from jinja2 import Template


class Replacer:
    def __init__(self, serializers):

        if isinstance(serializers, collections.MutableSequence):
            self.serializers = serializers
        else:
            self.serializers = [serializers]

    def run_all_serializers(self, current_version_dict, new_version_dict):
        summary = []
        for serializer in self.serializers:
            template = Template(serializer)

            summary.append((template.render(**current_version_dict), template.render(**new_version_dict)))

        return summary

    def run_main_serializer(self, current_version_dict, new_version_dict):
        return self.run_all_serializers(current_version_dict, new_version_dict)[0]

    def replace(self, text, current_version, new_version):
        new_text = text
        for serializer in self.serializers:
            template = Template(serializer)

            _search_pattern = template.render(**current_version)
            _replace_pattern = template.render(**new_version)

            new_text = new_text.replace(_search_pattern, _replace_pattern)

        return new_text



# -*- coding: utf-8 -*-

import os
import pytest

from punch import version as ver
from punch import version_part as vp


def clean_previous_imports():
    import sys

    for i in ['punch_config', 'punch_version']:
        if i in sys.modules:
            sys.modules.pop(i)


@pytest.fixture
def version_mmp():
    v = ver.Version()
    v.create_part('major', 4)
    v.create_part('minor', 3)
    v.create_part('patch', 1)
    return v


@pytest.fixture
def version_mmpb():
    v = ver.Version()
    v.create_part('major', 4)
    v.create_part('minor', 3)
    v.create_part('patch', 1)
    v.create_part('build', 5, start_value=1)
    return v


def test_version_default_part_is_integer():
    v = ver.Version()
    v.create_part('major', 4)
    assert isinstance(v.get_part('major'), vp.IntegerVersionPart)


def test_version_add_parts():
    v = ver.Version()
    part_major = vp.IntegerVersionPart('major', 4)
    part_minor = vp.IntegerVersionPart('minor', 3)
    v.add_part(part_major)
    v.add_part(part_minor)

    assert v.get_part('major').value == 4
    assert v.get_part('minor').value == 3
    assert v.keys == ['major', 'minor']


def test_version_may_specify_part_class():
    v = ver.Version()
    v.create_part('major', 4, vp.ValueListVersionPart, [0, 2, 4, 6, 8])
    assert isinstance(v.get_part('major'), vp.ValueListVersionPart)
    assert v.get_part('major').value == 4
    assert v.get_part('major').values == [0, 2, 4, 6, 8]


def test_version_can_add_parts(version_mmp):
    assert version_mmp.get_part('major').value == 4


def test_version_increment_last_part(version_mmp):
    version_mmp.inc('patch')
    assert version_mmp.get_part('patch').value == 2


def test_version_increment_middle_part(version_mmp):
    version_mmp.inc('minor')
    assert version_mmp.get_part('minor').value == 4
    assert version_mmp.get_part('patch').value == 0


def test_version_increment_first_part(version_mmp):
    version_mmp.inc('major')
    assert version_mmp.get_part('major').value == 5
    assert version_mmp.get_part('minor').value == 0
    assert version_mmp.get_part('patch').value == 0


def test_version_increment_part_with_custom_start_value(version_mmpb):
    version_mmpb.inc('major')
    assert version_mmpb.get_part('major').value == 5
    assert version_mmpb.get_part('minor').value == 0
    assert version_mmpb.get_part('patch').value == 0
    assert version_mmpb.get_part('build').value == 1


def test_version_copy(version_mmp):
    new_version = version_mmp.copy()
    new_version.inc('major')
    assert new_version.get_part('major').value == 5
    assert new_version.get_part('minor').value == 0
    assert new_version.get_part('patch').value == 0


def test_version_as_list(version_mmp):
    assert version_mmp.as_list() == [('major', 4), ('minor', 3), ('patch', 1)]


def test_version_keys_and_values(version_mmp):
    assert version_mmp.keys == ['major', 'minor', 'patch']
    assert [i.value for i in version_mmp.values] == [4, 3, 1]


def test_version_keys_keep_indertion_order(version_mmp):
    minor = version_mmp.parts.pop('minor')
    version_mmp.add_part(minor)
    assert version_mmp.as_list() == [('major', 4), ('patch', 1), ('minor', 3)]


def test_version_as_dict(version_mmp):
    expected_dict = {
        'major': 4,
        'minor': 3,
        'patch': 1
    }

    assert version_mmp.as_dict() == expected_dict


def test_write_version_file(temp_empty_dir, version_mmp):
    clean_previous_imports()

    version_filepath = os.path.join(temp_empty_dir, 'punch_version.py')

    version_mmp.to_file(version_filepath)

    with open(version_filepath, 'r') as f:
        content = sorted(f.readlines())

    expected_content = [
        "major = 4\n",
        "minor = 3\n",
        "patch = 1\n"
    ]

    assert content == expected_content


def test_read_complete_version_from_file(temp_empty_dir, version_mmp):
    clean_previous_imports()

    version_filepath = os.path.join(temp_empty_dir, 'punch_version.py')

    with open(version_filepath, 'w') as f:
        f.writelines(["major = 4\n", "minor = 3\n", "patch = 1\n"])

    version_description = [
        {
            'name': 'major',
            'type': 'integer'
        },
        {
            'name': 'minor',
            'type': 'integer'
        },
        {
            'name': 'patch',
            'type': 'integer'
        }
    ]

    version = ver.Version.from_file(version_filepath, version_description)

    assert version.keys == ['major', 'minor', 'patch']
    assert len(version.parts) == 3
    assert version.parts['major'].value == 4
    assert version.parts['minor'].value == 3
    assert version.parts['patch'].value == 1


def test_read_simplified_version_from_file(temp_empty_dir, version_mmp):
    clean_previous_imports()

    version_filepath = os.path.join(temp_empty_dir, 'punch_version.py')

    with open(version_filepath, 'w') as f:
        f.writelines(["major = 4\n", "minor = 3\n", "patch = 1\n"])

    version_description = ['major', 'minor', 'patch']

    version = ver.Version.from_file(version_filepath, version_description)

    assert version.keys == ['major', 'minor', 'patch']
    assert len(version.parts) == 3
    assert version.parts['major'].value == 4
    assert version.parts['minor'].value == 3
    assert version.parts['patch'].value == 1
